"""Tests for filo module."""

import filo
from pathlib import Path
from filo import Series

module_path = Path(filo.__file__).parent / '..'
data_path = module_path / 'data'

folders = data_path / 'img1', data_path / 'img2'
series = Series(folders, savepath=data_path, extension='.png')


def test_series_numbering():
    """Verify numbering of files is ok in multiple folders for series."""
    assert series.files[-1].num == 19


def test_series_info():
    """test generation of infos DataFrame."""
    series.load_info('External_File_Info.txt')
    info = series.info
    assert round(info.at[4, 'time (unix)']) == 1599832405


def test_series_info_overwrite():
    """Test loading file data from external file."""
    series.load_info('External_File_Info.txt')
    info = series.info
    assert info.at[2, 'time (unix)'] == 1599832401


def test_series_info_update_time():
    """Test loading file data from external file."""
    series.load_time('External_Time_Info.txt')
    info = series.info
    assert info.at[2, 'time (unix)'] == 1607500504


def test_series_duration():
    """Test calculation of time duration of series."""
    series.load_time('External_Time_Info.txt')
    assert round(series.duration.total_seconds()) == 38



"""Manage Series of experimental data in potentially several folders."""

import os
from pathlib import Path
import datetime

import pandas as pd

from .general import make_iterable, list_files


# ================================= Classes ==================================

class File:
    """Individual file among the series of files. Used by Series"""

    def __init__(self, file, num):
        """Parameters:

         - file: pathlib object linking to the file
         - num: number identifier of the file across all folders
         """
        self.file = file            # Pathlib object
        self.num = num              # identifier (int)
        self.time = None  # stores time when Series._measure_times() is called

    def __repr__(self):
        return f"filo.File #{self.num} [{self.name} in folder '{self.folder}']"

    @property
    def folder(self):
        return self.file.parent

    @property
    def name(self):
        return self.file.name


class Series:
    """Class to manage series of experimental data in one or several folders."""

    name = 'File series'
    info_filename = 'FileSeries_Info.txt'

    def __init__(self, paths='.', extension='', savepath='.'):
        """Init series of spectra object.

        Parameters
        ----------
        - paths can be a string, path object, or a list of str/paths if data
          is stored in multiple folders.
        - extension: extension of files to be considered (e.g. '.txt')
        - savepath: folder in which to save analysis data if applicable.
        """
        self.savepath = Path(savepath)
        self.savepath.mkdir(exist_ok=True)  # create savepath if does not exist.

        if type(paths) is str:
            self.folders = paths,
        else:
            self.folders = make_iterable(paths)
        self.folders = [Path(folder) for folder in self.folders]
        self.extension = extension

        self.times_set = False  # True when file times are set (e.g. by measure_time())

        self.files = self._list_files()  # empty list if no files in folder.
        self._measure_times()             # add creation time info

    def __repr__(self):
        return f"{self.__class__.name} [extension '{self.extension}', "\
               f'folders {[str(folder) for folder in self.folders]}, '\
               f"savepath '{self.savepath}', {len(self.files)} files]"

    # =================== Private methods to get files info ==================

    def _list_files(self):
        """Create list of filo.File objects from files in self.folders.

        Return a list of filo.File objects.
        """
        files = []
        n = 0  # shift in measurement number if multiple folders
        for i_folder, folder in enumerate(self.folders):
            for i_local, file in enumerate(list_files(folder, self.extension)):
                num = i_local + n
                files.append(File(file, num))
            else:
                try:
                    n += i_local + 1
                except UnboundLocalError:  # no data in folder
                    pass
        return files

    def _measure_times(self):
        """Define time (unix) associated with each file (subclass if necessary)."""
        for file in self.files:
            file.time = file.file.stat().st_mtime
        self.times_set = True

    # =========================== Public methods =============================

    @property
    def info(self):
        """Get time of files from creation date (st_mtime).

        Output
        ------
        pandas dataframe with 'num' as index and 'time (unix)', folder, filename
        as columns.
        """
        nums = [file.num for file in self.files]
        filenames = [file.name for file in self.files]
        dirs = [os.path.relpath(f.folder, self.savepath) for f in self.files]

        columns = ['num', 'folder', 'filename']
        info = zip(nums, dirs, filenames)
        data = pd.DataFrame(info, columns=columns)
        data.set_index('num', inplace=True)

        if self.times_set:
            times = [file.time for file in self.files]
            data['time (unix)'] = times

        return data

    def save_info(self, filename=None, sep='\t'):
        """Save info DataFrame (see self.info property) into csv file.

        Use default filename if filename is None, else specified name.
        The file is created in self.savepath (default current directory).
        """
        filename = self.info_filename if filename is None else filename
        data = self.info  # pandas dataframe (see self.times below)
        time_file = Path(self.savepath) / filename
        data.to_csv(time_file, sep=sep)

    def load_info(self, filename=None, sep='\t'):
        """Update files list and infos DataFrame using csv-saved data.

        Data must be saved in a csv file in self.savepath, with csv separator
        `sep`, and with columns (at least) 'num', 'folder', 'filename'.
        If timing info is available (column 'time (unix)'), file.time is set
        as well.

        No output (self.files and self.info are updated).
        """
        filename = self.info_filename if filename is None else filename
        data_file = Path(self.savepath) / filename
        data = pd.read_csv(data_file, sep=sep)
        data.set_index('num', inplace=True)

        # Check if there is time info ----------------------------------------
        try:
            data['time (unix)']
        except KeyError:
            self.times_set = False
        else:
            self.times_set = True

        # Update file information in the list of filo.File objects -----------
        self.files = []
        for num in data.index:
            filename = data.at[num, 'filename']
            foldername = data.at[num, 'folder']
            filepath = self.savepath / foldername / filename
            file = File(filepath, num)
            if self.times_set:
                file.time = data.at[file.num, 'time (unix)']
            self.files.append(file)

    def load_time(self, filename=None, sep='\t'):
        """Same as load_files but updates only times, not file info.

        The data in filename must be csv with as least columns 'num' and
        'time (unix)'. All nums in self.files must be present in the data.
        """
        filename = self.info_filename if filename is None else filename
        data_file = Path(self.savepath) / filename
        data = pd.read_csv(data_file, sep=sep)
        data.set_index('num', inplace=True)

        # Update file information in the list of filo.File objects -----------
        for file in self.files:
            file.time = data.at[file.num, 'time (unix)']

    @property
    def duration(self):
        """Timedelta between timing info of first and last file.

        Output
        ------
        datetime.Timedelta object.
        """
        if self.times_set:
            t = 'time (unix)'
            dt_s = self.info[t].iloc[-1] - self.info[t].iloc[0]
            return datetime.timedelta(seconds=float(dt_s))
        else:
            raise AttributeError('File timing info missing.')



"""File Management."""

from .general import list_files, list_all, move_files, move_all
from .general import batch_file_rename, data_to_line, line_to_data
from .general import make_iterable

from .series import Series

from importlib.metadata import version

__version__ = version('filo')



"""File Management."""


from pathlib import Path


# ========================== Basic File Management ===========================


def list_files(path='.', extension=''):
    """Return list of files with specific extension in path.

    - by default, return all files with any extension.
    - directories are excluded
    - results are sorted by name
    """
    folder = Path(path)
    pattern = '*' + extension
    paths = folder.glob(pattern)
    files = [p for p in paths if p.is_file()]
    return sorted(files)


def list_all(path='.'):
    """List all contents of a folder, sorted by name."""
    folder = Path(path)
    contents = folder.glob('*')
    return sorted(contents)


def move_files(src='.', dst='.', extension=''):
    """Move all files with a certain extension from folder 1 to folder 2.

    - directories are excluded
    - directory dst is created if not already existing
    """
    p1, p2 = Path(src), Path(dst)
    files = list_files(p1, extension)
    p2.mkdir(exist_ok=True)
    for file in files:
        newfile = p2 / file.name
        file.rename(newfile)


def move_all(src='.', dst='.'):
    """Move all contents of folder 1 into folder 2.

    directory dst is created if not already existing
    """
    p1, p2 = Path(src), Path(dst)
    p2.mkdir(exist_ok=True)
    contents = p1.glob('*')  # I do not use list_all to keep a generator
    for content in contents:
        new_content = p2 / content.name
        content.rename(new_content)


def batch_file_rename(name, newname, path='.'):
    """Change name to newname for all files in path and subdirectories.

    Parameters
    ----------
    name, newname: str
    path: str or path object

    Example
    -------
    batch_file_rename('Test.txt', 'test.txt')
    will rename all files named 'Test.txt' into 'test.txt' in current directory
    and subdirectories.

    Notes
    -----
    - IMPORTANT: before running this function, check all files that will be
    impacted by running list(Path(path).glob(f'**/{name}')).
    - modification, creation dates etc. not changed in the process.
    """
    folder = Path(path)
    files = folder.glob(f'**/{name}')

    for file in files:
        newfile = file.with_name(newname)
        file.rename(newfile)


# =================== Functions for csv saving and reading ===================


def data_to_line(data, sep='\t'):
    """Transform iterable into line to write in a file, with a separarator."""
    data_str_list = [str(x) for x in data]
    data_str_all = sep.join(data_str_list)
    return data_str_all + '\n'


def line_to_data(line, sep='\t', dtype=float):
    """Inverse of data_to_line(). Returns data as a tuple of type dtype."""
    line_list = line.split(sep)
    data_list = [dtype(x) for x in line_list]
    return tuple(data_list)


# ================================== MISC. ===================================


def make_iterable(x):
    """Transform non-iterables into a tuple, but keeps iterables unchanged."""
    try:
        iter(x)
    except TypeError:
        return x,
    else:
        return x
